@@ -6,6 +6,7 @@
 
 #include "curve25519.h"
 
+#include <linux/version.h>
 #include <linux/string.h>
 #include <linux/random.h>
 #include <crypto/algapi.h>
@@ -29,32 +30,112 @@ static const u8 null_point[CURVE25519_POINT_SIZE] = { 0 };
 #include <asm/processor.h>
 #include <asm/fpu/api.h>
 #include <asm/simd.h>
-static bool curve25519_use_avx __read_mostly = false;
+static bool curve25519_use_avx __read_mostly;
 void curve25519_fpu_init(void)
 {
-	curve25519_use_avx = boot_cpu_has(X86_FEATURE_AVX);
+	curve25519_use_avx = boot_cpu_has(X86_FEATURE_AVX) && cpu_has_xfeatures(XFEATURE_MASK_SSE | XFEATURE_MASK_YMM, NULL);
 }
 
 typedef u64 fe[10];
 typedef u64 fe51[5];
 asmlinkage void curve25519_sandy2x_ladder(fe *, const u8 *);
 asmlinkage void curve25519_sandy2x_ladder_base(fe *, const u8 *);
-asmlinkage void curve25519_sandy2x_fe_frombytes(fe, const u8 *);
 asmlinkage void curve25519_sandy2x_fe51_pack(u8 *, const fe51 *);
 asmlinkage void curve25519_sandy2x_fe51_mul(fe51 *, const fe51 *, const fe51 *);
-asmlinkage void curve25519_sandy2x_fe51_invert(fe51 *, const fe51 *);
+asmlinkage void curve25519_sandy2x_fe51_nsquare(fe51 *, const fe51 *, int);
+
+static inline u32 le24_to_cpupv(const u8 *in)
+{
+	return le16_to_cpup((__le16 *)in) | ((u32)in[2]) << 16;
+}
+
+static inline void fe_frombytes(fe h, const u8 *s)
+{
+	u64 h0 = le32_to_cpup((__le32 *)s);
+	u64 h1 = le24_to_cpupv(s + 4) << 6;
+	u64 h2 = le24_to_cpupv(s + 7) << 5;
+	u64 h3 = le24_to_cpupv(s + 10) << 3;
+	u64 h4 = le24_to_cpupv(s + 13) << 2;
+	u64 h5 = le32_to_cpup((__le32 *)(s + 16));
+	u64 h6 = le24_to_cpupv(s + 20) << 7;
+	u64 h7 = le24_to_cpupv(s + 23) << 5;
+	u64 h8 = le24_to_cpupv(s + 26) << 4;
+	u64 h9 = (le24_to_cpupv(s + 29) & 8388607) << 2;
+	u64 carry0, carry1, carry2, carry3, carry4, carry5, carry6, carry7, carry8, carry9;
+
+	carry9 = h9 >> 25; h0 += carry9 * 19; h9 &= 0x1FFFFFF;
+	carry1 = h1 >> 25; h2 += carry1; h1 &= 0x1FFFFFF;
+	carry3 = h3 >> 25; h4 += carry3; h3 &= 0x1FFFFFF;
+	carry5 = h5 >> 25; h6 += carry5; h5 &= 0x1FFFFFF;
+	carry7 = h7 >> 25; h8 += carry7; h7 &= 0x1FFFFFF;
+
+	carry0 = h0 >> 26; h1 += carry0; h0 &= 0x3FFFFFF;
+	carry2 = h2 >> 26; h3 += carry2; h2 &= 0x3FFFFFF;
+	carry4 = h4 >> 26; h5 += carry4; h4 &= 0x3FFFFFF;
+	carry6 = h6 >> 26; h7 += carry6; h6 &= 0x3FFFFFF;
+	carry8 = h8 >> 26; h9 += carry8; h8 &= 0x3FFFFFF;
+
+	h[0] = h0;
+	h[1] = h1;
+	h[2] = h2;
+	h[3] = h3;
+	h[4] = h4;
+	h[5] = h5;
+	h[6] = h6;
+	h[7] = h7;
+	h[8] = h8;
+	h[9] = h9;
+}
+
+static inline void fe51_invert(fe51 *r, const fe51 *x)
+{
+	fe51 z2, z9, z11, z2_5_0, z2_10_0, z2_20_0, z2_50_0, z2_100_0, t;
+
+	/* 2 */ curve25519_sandy2x_fe51_nsquare(&z2, x, 1);
+	/* 4 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&z2, 1);
+	/* 8 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&t, 1);
+	/* 9 */ curve25519_sandy2x_fe51_mul(&z9, (const fe51 *)&t, x);
+	/* 11 */ curve25519_sandy2x_fe51_mul(&z11, (const fe51 *)&z9, (const fe51 *)&z2);
+	/* 22 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&z11, 1);
+	/* 2^5 - 2^0 = 31 */ curve25519_sandy2x_fe51_mul(&z2_5_0, (const fe51 *)&t, (const fe51 *)&z9);
+
+	/* 2^10 - 2^5 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&z2_5_0, 5);
+	/* 2^10 - 2^0 */ curve25519_sandy2x_fe51_mul(&z2_10_0, (const fe51 *)&t, (const fe51 *)&z2_5_0);
+
+	/* 2^20 - 2^10 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&z2_10_0, 10);
+	/* 2^20 - 2^0 */ curve25519_sandy2x_fe51_mul(&z2_20_0, (const fe51 *)&t, (const fe51 *)&z2_10_0);
+
+	/* 2^40 - 2^20 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&z2_20_0, 20);
+	/* 2^40 - 2^0 */ curve25519_sandy2x_fe51_mul(&t, (const fe51 *)&t, (const fe51 *)&z2_20_0);
+
+	/* 2^50 - 2^10 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&t, 10);
+	/* 2^50 - 2^0 */ curve25519_sandy2x_fe51_mul(&z2_50_0, (const fe51 *)&t, (const fe51 *)&z2_10_0);
+
+	/* 2^100 - 2^50 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&z2_50_0, 50);
+	/* 2^100 - 2^0 */ curve25519_sandy2x_fe51_mul(&z2_100_0, (const fe51 *)&t, (const fe51 *)&z2_50_0);
+
+	/* 2^200 - 2^100 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&z2_100_0, 100);
+	/* 2^200 - 2^0 */ curve25519_sandy2x_fe51_mul(&t, (const fe51 *)&t, (const fe51 *)&z2_100_0);
+
+	/* 2^250 - 2^50 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&t, 50);
+	/* 2^250 - 2^0 */ curve25519_sandy2x_fe51_mul(&t, (const fe51 *)&t, (const fe51 *)&z2_50_0);
+
+	/* 2^255 - 2^5 */ curve25519_sandy2x_fe51_nsquare(&t, (const fe51 *)&t, 5);
+	/* 2^255 - 21 */ curve25519_sandy2x_fe51_mul(r, (const fe51 *)t, (const fe51 *)&z11);
+}
 
 static void curve25519_sandy2x(u8 mypublic[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_POINT_SIZE], const u8 basepoint[CURVE25519_POINT_SIZE])
 {
 	u8 e[32];
 	fe var[3];
 	fe51 x_51, z_51;
+
 	memcpy(e, secret, 32);
 	normalize_secret(e);
 #define x1 var[0]
 #define x2 var[1]
 #define z2 var[2]
-	curve25519_sandy2x_fe_frombytes(x1, basepoint);
+	fe_frombytes(x1, basepoint);
 	curve25519_sandy2x_ladder(var, e);
 	z_51[0] = (z2[1] << 26) + z2[0];
 	z_51[1] = (z2[3] << 26) + z2[2];
@@ -69,7 +150,7 @@ static void curve25519_sandy2x(u8 mypublic[CURVE25519_POINT_SIZE], const u8 secr
 #undef x1
 #undef x2
 #undef z2
-	curve25519_sandy2x_fe51_invert(&z_51, (const fe51 *)&z_51);
+	fe51_invert(&z_51, (const fe51 *)&z_51);
 	curve25519_sandy2x_fe51_mul(&x_51, (const fe51 *)&x_51, (const fe51 *)&z_51);
 	curve25519_sandy2x_fe51_pack(mypublic, (const fe51 *)&x_51);
 
@@ -84,6 +165,7 @@ static void curve25519_sandy2x_base(u8 pub[CURVE25519_POINT_SIZE], const u8 secr
 	u8 e[32];
 	fe var[3];
 	fe51 x_51, z_51;
+
 	memcpy(e, secret, 32);
 	normalize_secret(e);
 	curve25519_sandy2x_ladder_base(var, e);
@@ -101,7 +183,7 @@ static void curve25519_sandy2x_base(u8 pub[CURVE25519_POINT_SIZE], const u8 secr
 	x_51[4] = (x2[9] << 26) + x2[8];
 #undef x2
 #undef z2
-	curve25519_sandy2x_fe51_invert(&z_51, (const fe51 *)&z_51);
+	fe51_invert(&z_51, (const fe51 *)&z_51);
 	curve25519_sandy2x_fe51_mul(&x_51, (const fe51 *)&x_51, (const fe51 *)&z_51);
 	curve25519_sandy2x_fe51_pack(pub, (const fe51 *)&x_51);
 
@@ -110,24 +192,24 @@ static void curve25519_sandy2x_base(u8 pub[CURVE25519_POINT_SIZE], const u8 secr
 	memzero_explicit(x_51, sizeof(x_51));
 	memzero_explicit(z_51, sizeof(z_51));
 }
-#elif IS_ENABLED(CONFIG_KERNEL_MODE_NEON) && defined(CONFIG_ARM)
+#elif IS_ENABLED(CONFIG_KERNEL_MODE_NEON) && defined(CONFIG_ARM) && !defined(CONFIG_CPU_THUMBONLY)
 #include <asm/hwcap.h>
 #include <asm/neon.h>
 #include <asm/simd.h>
 asmlinkage void curve25519_asm_neon(u8 mypublic[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_POINT_SIZE], const u8 basepoint[CURVE25519_POINT_SIZE]);
-static bool curve25519_use_neon __read_mostly = false;
-void curve25519_fpu_init(void)
+static bool curve25519_use_neon __read_mostly;
+void __init curve25519_fpu_init(void)
 {
 	curve25519_use_neon = elf_hwcap & HWCAP_NEON;
 }
 #else
-void curve25519_fpu_init(void) { }
+void __init curve25519_fpu_init(void) { }
 #endif
 
-#ifdef __SIZEOF_INT128__
+#if defined(CONFIG_ARCH_SUPPORTS_INT128) && defined(__SIZEOF_INT128__)
 typedef u64 limb;
 typedef limb felem[5];
-typedef __uint128_t uint128_t;
+typedef __uint128_t u128;
 
 /* Sum two numbers: output += in */
 static __always_inline void fsum(limb *output, const limb *in)
@@ -161,21 +243,21 @@ static __always_inline void fdifference_backwards(felem out, const felem in)
 /* Multiply a number by a scalar: output = in * scalar */
 static __always_inline void fscalar_product(felem output, const felem in, const limb scalar)
 {
-	uint128_t a;
+	u128 a;
 
-	a = ((uint128_t) in[0]) * scalar;
+	a = ((u128) in[0]) * scalar;
 	output[0] = ((limb)a) & 0x7ffffffffffffUL;
 
-	a = ((uint128_t) in[1]) * scalar + ((limb) (a >> 51));
+	a = ((u128) in[1]) * scalar + ((limb) (a >> 51));
 	output[1] = ((limb)a) & 0x7ffffffffffffUL;
 
-	a = ((uint128_t) in[2]) * scalar + ((limb) (a >> 51));
+	a = ((u128) in[2]) * scalar + ((limb) (a >> 51));
 	output[2] = ((limb)a) & 0x7ffffffffffffUL;
 
-	a = ((uint128_t) in[3]) * scalar + ((limb) (a >> 51));
+	a = ((u128) in[3]) * scalar + ((limb) (a >> 51));
 	output[3] = ((limb)a) & 0x7ffffffffffffUL;
 
-	a = ((uint128_t) in[4]) * scalar + ((limb) (a >> 51));
+	a = ((u128) in[4]) * scalar + ((limb) (a >> 51));
 	output[4] = ((limb)a) & 0x7ffffffffffffUL;
 
 	output[0] += (a >> 51) * 19;
@@ -191,8 +273,8 @@ static __always_inline void fscalar_product(felem output, const felem in, const
  */
 static __always_inline void fmul(felem output, const felem in2, const felem in)
 {
-	uint128_t t[5];
-	limb r0,r1,r2,r3,r4,s0,s1,s2,s3,s4,c;
+	u128 t[5];
+	limb r0, r1, r2, r3, r4, s0, s1, s2, s3, s4, c;
 
 	r0 = in[0];
 	r1 = in[1];
@@ -206,21 +288,21 @@ static __always_inline void fmul(felem output, const felem in2, const felem in)
 	s3 = in2[3];
 	s4 = in2[4];
 
-	t[0]  =  ((uint128_t) r0) * s0;
-	t[1]  =  ((uint128_t) r0) * s1 + ((uint128_t) r1) * s0;
-	t[2]  =  ((uint128_t) r0) * s2 + ((uint128_t) r2) * s0 + ((uint128_t) r1) * s1;
-	t[3]  =  ((uint128_t) r0) * s3 + ((uint128_t) r3) * s0 + ((uint128_t) r1) * s2 + ((uint128_t) r2) * s1;
-	t[4]  =  ((uint128_t) r0) * s4 + ((uint128_t) r4) * s0 + ((uint128_t) r3) * s1 + ((uint128_t) r1) * s3 + ((uint128_t) r2) * s2;
+	t[0]  =  ((u128) r0) * s0;
+	t[1]  =  ((u128) r0) * s1 + ((u128) r1) * s0;
+	t[2]  =  ((u128) r0) * s2 + ((u128) r2) * s0 + ((u128) r1) * s1;
+	t[3]  =  ((u128) r0) * s3 + ((u128) r3) * s0 + ((u128) r1) * s2 + ((u128) r2) * s1;
+	t[4]  =  ((u128) r0) * s4 + ((u128) r4) * s0 + ((u128) r3) * s1 + ((u128) r1) * s3 + ((u128) r2) * s2;
 
 	r4 *= 19;
 	r1 *= 19;
 	r2 *= 19;
 	r3 *= 19;
 
-	t[0] += ((uint128_t) r4) * s1 + ((uint128_t) r1) * s4 + ((uint128_t) r2) * s3 + ((uint128_t) r3) * s2;
-	t[1] += ((uint128_t) r4) * s2 + ((uint128_t) r2) * s4 + ((uint128_t) r3) * s3;
-	t[2] += ((uint128_t) r4) * s3 + ((uint128_t) r3) * s4;
-	t[3] += ((uint128_t) r4) * s4;
+	t[0] += ((u128) r4) * s1 + ((u128) r1) * s4 + ((u128) r2) * s3 + ((u128) r3) * s2;
+	t[1] += ((u128) r4) * s2 + ((u128) r2) * s4 + ((u128) r3) * s3;
+	t[2] += ((u128) r4) * s3 + ((u128) r3) * s4;
+	t[3] += ((u128) r4) * s4;
 
 			r0 = (limb)t[0] & 0x7ffffffffffffUL; c = (limb)(t[0] >> 51);
 	t[1] += c;      r1 = (limb)t[1] & 0x7ffffffffffffUL; c = (limb)(t[1] >> 51);
@@ -240,9 +322,9 @@ static __always_inline void fmul(felem output, const felem in2, const felem in)
 
 static __always_inline void fsquare_times(felem output, const felem in, limb count)
 {
-	uint128_t t[5];
-	limb r0,r1,r2,r3,r4,c;
-	limb d0,d1,d2,d4,d419;
+	u128 t[5];
+	limb r0, r1, r2, r3, r4, c;
+	limb d0, d1, d2, d4, d419;
 
 	r0 = in[0];
 	r1 = in[1];
@@ -257,11 +339,11 @@ static __always_inline void fsquare_times(felem output, const felem in, limb cou
 		d419 = r4 * 19;
 		d4 = d419 * 2;
 
-		t[0] = ((uint128_t) r0) * r0 + ((uint128_t) d4) * r1 + (((uint128_t) d2) * (r3     ));
-		t[1] = ((uint128_t) d0) * r1 + ((uint128_t) d4) * r2 + (((uint128_t) r3) * (r3 * 19));
-		t[2] = ((uint128_t) d0) * r2 + ((uint128_t) r1) * r1 + (((uint128_t) d4) * (r3     ));
-		t[3] = ((uint128_t) d0) * r3 + ((uint128_t) d1) * r2 + (((uint128_t) r4) * (d419   ));
-		t[4] = ((uint128_t) d0) * r4 + ((uint128_t) d1) * r3 + (((uint128_t) r2) * (r2     ));
+		t[0] = ((u128) r0) * r0 + ((u128) d4) * r1 + (((u128) d2) * (r3     ));
+		t[1] = ((u128) d0) * r1 + ((u128) d4) * r2 + (((u128) r3) * (r3 * 19));
+		t[2] = ((u128) d0) * r2 + ((u128) r1) * r1 + (((u128) d4) * (r3     ));
+		t[3] = ((u128) d0) * r3 + ((u128) d1) * r2 + (((u128) r4) * (d419   ));
+		t[4] = ((u128) d0) * r4 + ((u128) d1) * r3 + (((u128) r2) * (r2     ));
 
 				r0 = (limb)t[0] & 0x7ffffffffffffUL; c = (limb)(t[0] >> 51);
 		t[1] += c;      r1 = (limb)t[1] & 0x7ffffffffffffUL; c = (limb)(t[1] >> 51);
@@ -271,7 +353,7 @@ static __always_inline void fsquare_times(felem output, const felem in, limb cou
 		r0 +=   c * 19; c = r0 >> 51; r0 = r0 & 0x7ffffffffffffUL;
 		r1 +=   c;      c = r1 >> 51; r1 = r1 & 0x7ffffffffffffUL;
 		r2 +=   c;
-	} while(--count);
+	} while (--count);
 
 	output[0] = r0;
 	output[1] = r1;
@@ -283,12 +365,12 @@ static __always_inline void fsquare_times(felem output, const felem in, limb cou
 /* Load a little-endian 64-bit number  */
 static inline limb load_limb(const u8 *in)
 {
-	return le64_to_cpu(*(u64 *)in);
+	return le64_to_cpu(*(__le64 *)in);
 }
 
 static inline void store_limb(u8 *out, limb in)
 {
-	*(u64 *)out = cpu_to_le64(in);
+	*(__le64 *)out = cpu_to_le64(in);
 }
 
 /* Take a little-endian, 32-byte number and expand it into polynomial form */
@@ -306,7 +388,7 @@ static inline void fexpand(limb *output, const u8 *in)
  */
 static void fcontract(u8 *output, const felem input)
 {
-	uint128_t t[5];
+	u128 t[5];
 
 	t[0] = input[0];
 	t[1] = input[1];
@@ -372,6 +454,7 @@ static void fmonty(limb *x2, limb *z2, /* output 2Q */
 			 limb *x3, limb *z3, /* output Q + Q' */
 			 limb *x, limb *z,   /* input Q */
 			 limb *xprime, limb *zprime, /* input Q' */
+
 			 const limb *qmqp /* input Q - Q' */)
 {
 	limb origx[5], origxprime[5], zzz[5], xx[5], zz[5], xxprime[5], zzprime[5], zzzprime[5];
@@ -409,11 +492,12 @@ static void fmonty(limb *x2, limb *z2, /* output 2Q */
  */
 static void swap_conditional(limb a[5], limb b[5], limb iswap)
 {
-	unsigned i;
+	unsigned int i;
 	const limb swap = -iswap;
 
 	for (i = 0; i < 5; ++i) {
 		const limb x = swap & (a[i] ^ b[i]);
+
 		a[i] ^= x;
 		b[i] ^= x;
 	}
@@ -432,12 +516,13 @@ static void cmult(limb *resultx, limb *resultz, const u8 *n, const limb *q)
 	limb e[5] = {0}, f[5] = {1}, g[5] = {0}, h[5] = {1};
 	limb *nqpqx2 = e, *nqpqz2 = f, *nqx2 = g, *nqz2 = h;
 
-	unsigned i, j;
+	unsigned int i, j;
 
 	memcpy(nqpqx, q, sizeof(limb) * 5);
 
 	for (i = 0; i < 32; ++i) {
 		u8 byte = n[31 - i];
+
 		for (j = 0; j < 8; ++j) {
 			const limb bit = byte >> 7;
 
@@ -474,7 +559,7 @@ static void cmult(limb *resultx, limb *resultz, const u8 *n, const limb *q)
 
 static void crecip(felem out, const felem z)
 {
-	felem a,t0,b,c;
+	felem a, t0, b, c;
 
 	/* 2 */ fsquare_times(a, z, 1); // a = 2
 	/* 8 */ fsquare_times(t0, a, 2);
@@ -534,6 +619,10 @@ bool curve25519(u8 mypublic[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_P
 bool curve25519_generate_public(u8 pub[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_POINT_SIZE])
 {
 	static const u8 basepoint[CURVE25519_POINT_SIZE] __aligned(32) = { 9 };
+
+	if (unlikely(!crypto_memneq(secret, null_point, CURVE25519_POINT_SIZE)))
+		return false;
+
 #ifdef CONFIG_X86_64
 	if (curve25519_use_avx && irq_fpu_usable()) {
 		kernel_fpu_begin();
@@ -553,12 +642,14 @@ typedef s64 limb;
  * significant first. The value of the field element is:
  *   x[0] + 2^26·x[1] + x^51·x[2] + 2^102·x[3] + ...
  *
- * i.e. the limbs are 26, 25, 26, 25, ... bits wide. */
+ * i.e. the limbs are 26, 25, 26, 25, ... bits wide.
+ */
 
 /* Sum two numbers: output += in */
 static void fsum(limb *output, const limb *in)
 {
-	unsigned i;
+	unsigned int i;
+
 	for (i = 0; i < 10; i += 2) {
 		output[0 + i] = output[0 + i] + in[0 + i];
 		output[1 + i] = output[1 + i] + in[1 + i];
@@ -566,10 +657,12 @@ static void fsum(limb *output, const limb *in)
 }
 
 /* Find the difference of two numbers: output = in - output
- * (note the order of the arguments!). */
+ * (note the order of the arguments!).
+ */
 static void fdifference(limb *output, const limb *in)
 {
-	unsigned i;
+	unsigned int i;
+
 	for (i = 0; i < 10; ++i)
 		output[i] = in[i] - output[i];
 }
@@ -577,7 +670,8 @@ static void fdifference(limb *output, const limb *in)
 /* Multiply a number by a scalar: output = in * scalar */
 static void fscalar_product(limb *output, const limb *in, const limb scalar)
 {
-	unsigned i;
+	unsigned int i;
+
 	for (i = 0; i < 10; ++i)
 		output[i] = in[i] * scalar;
 }
@@ -587,7 +681,8 @@ static void fscalar_product(limb *output, const limb *in, const limb scalar)
  * output must be distinct to both inputs. The inputs are reduced coefficient
  * form, the output is not.
  *
- * output[x] <= 14 * the largest product of the input limbs. */
+ * output[x] <= 14 * the largest product of the input limbs.
+ */
 static void fproduct(limb *output, const limb *in2, const limb *in)
 {
 	output[0] =       ((limb) ((s32) in2[0])) * ((s32) in[0]);
@@ -695,13 +790,15 @@ static void fproduct(limb *output, const limb *in2, const limb *in)
 /* Reduce a long form to a short form by taking the input mod 2^255 - 19.
  *
  * On entry: |output[i]| < 14*2^54
- * On exit: |output[0..8]| < 280*2^54 */
+ * On exit: |output[0..8]| < 280*2^54
+ */
 static void freduce_degree(limb *output)
 {
 	/* Each of these shifts and adds ends up multiplying the value by 19.
 	 *
 	 * For output[0..8], the absolute entry value is < 14*2^54 and we add, at
-	 * most, 19*14*2^54 thus, on exit, |output[0..8]| < 280*2^54. */
+	 * most, 19*14*2^54 thus, on exit, |output[0..8]| < 280*2^54.
+	 */
 	output[8] += output[18] << 4;
 	output[8] += output[18] << 1;
 	output[8] += output[18];
@@ -737,7 +834,8 @@ static void freduce_degree(limb *output)
 
 /* return v / 2^26, using only shifts and adds.
  *
- * On entry: v can take any value. */
+ * On entry: v can take any value.
+ */
 static inline limb div_by_2_26(const limb v)
 {
 	/* High word of v; no shift needed. */
@@ -752,7 +850,8 @@ static inline limb div_by_2_26(const limb v)
 
 /* return v / (2^25), using only shifts and adds.
  *
- * On entry: v can take any value. */
+ * On entry: v can take any value.
+ */
 static inline limb div_by_2_25(const limb v)
 {
 	/* High word of v; no shift needed*/
@@ -767,10 +866,11 @@ static inline limb div_by_2_25(const limb v)
 
 /* Reduce all coefficients of the short form input so that |x| < 2^26.
  *
- * On entry: |output[i]| < 280*2^54 */
+ * On entry: |output[i]| < 280*2^54
+ */
 static void freduce_coefficients(limb *output)
 {
-	unsigned i;
+	unsigned int i;
 
 	output[10] = 0;
 
@@ -779,7 +879,8 @@ static void freduce_coefficients(limb *output)
 		/* The entry condition (that |output[i]| < 280*2^54) means that over is, at
 		 * most, 280*2^28 in the first iteration of this loop. This is added to the
 		 * next limb and we can approximate the resulting bound of that limb by
-		 * 281*2^54. */
+		 * 281*2^54.
+		 */
 		output[i] -= over << 26;
 		output[i+1] += over;
 
@@ -788,7 +889,8 @@ static void freduce_coefficients(limb *output)
 		 * be approximated as 281*2^54.
 		 *
 		 * For subsequent iterations of the loop, 281*2^54 remains a conservative
-		 * bound and no overflow occurs. */
+		 * bound and no overflow occurs.
+		 */
 		over = div_by_2_25(output[i+1]);
 		output[i+1] -= over << 25;
 		output[i+2] += over;
@@ -801,15 +903,18 @@ static void freduce_coefficients(limb *output)
 	output[10] = 0;
 
 	/* Now output[1..9] are reduced, and |output[0]| < 2^26 + 19*281*2^29
-	 * So |over| will be no more than 2^16. */
+	 * So |over| will be no more than 2^16.
+	 */
 	{
 		limb over = div_by_2_26(output[0]);
+
 		output[0] -= over << 26;
 		output[1] += over;
 	}
 
 	/* Now output[0,2..9] are reduced, and |output[1]| < 2^25 + 2^16 < 2^26. The
-	 * bound on |output[1]| is sufficient to meet our needs. */
+	 * bound on |output[1]| is sufficient to meet our needs.
+	 */
 }
 
 /* A helpful wrapper around fproduct: output = in * in2.
@@ -817,10 +922,12 @@ static void freduce_coefficients(limb *output)
  * On entry: |in[i]| < 2^27 and |in2[i]| < 2^27.
  *
  * output must be distinct to both inputs. The output is reduced degree
- * (indeed, one need only provide storage for 10 limbs) and |output[i]| < 2^26. */
+ * (indeed, one need only provide storage for 10 limbs) and |output[i]| < 2^26.
+ */
 static void fmul(limb *output, const limb *in, const limb *in2)
 {
 	limb t[19];
+
 	fproduct(t, in, in2);
 	/* |t[i]| < 14*2^54 */
 	freduce_degree(t);
@@ -834,7 +941,8 @@ static void fmul(limb *output, const limb *in, const limb *in2)
  * output must be distinct from the input. The inputs are reduced coefficient
  * form, the output is not.
  *
- * output[x] <= 14 * the largest product of the input limbs. */
+ * output[x] <= 14 * the largest product of the input limbs.
+ */
 static void fsquare_inner(limb *output, const limb *in)
 {
 	output[0] =       ((limb) ((s32) in[0])) * ((s32) in[0]);
@@ -900,14 +1008,17 @@ static void fsquare_inner(limb *output, const limb *in)
  * 2^27.
  *
  * On exit: The |output| argument is in reduced coefficients form (indeed, one
- * need only provide storage for 10 limbs) and |out[i]| < 2^26. */
+ * need only provide storage for 10 limbs) and |out[i]| < 2^26.
+ */
 static void fsquare(limb *output, const limb *in)
 {
 	limb t[19];
+
 	fsquare_inner(t, in);
 	/* |t[i]| < 14*2^54 because the largest product of two limbs will be <
 	 * 2^(27+27) and fsquare_inner adds together, at most, 14 of those
-	 * products. */
+	 * products.
+	 */
 	freduce_degree(t);
 	freduce_coefficients(t);
 	/* |t[i]| < 2^26 */
@@ -917,7 +1028,7 @@ static void fsquare(limb *output, const limb *in)
 /* Take a little-endian, 32-byte number and expand it into polynomial form */
 static inline void fexpand(limb *output, const u8 *input)
 {
-#define F(n,start,shift,mask) \
+#define F(n, start, shift, mask) \
 	output[n] = ((((limb) input[start + 0]) | \
 		      ((limb) input[start + 1]) << 8 | \
 		      ((limb) input[start + 2]) << 16 | \
@@ -952,7 +1063,8 @@ static s32 s32_eq(s32 a, s32 b)
 }
 
 /* s32_gte returns 0xffffffff if a >= b and zero otherwise, where a and b are
- * both non-negative. */
+ * both non-negative.
+ */
 static s32 s32_gte(s32 a, s32 b)
 {
 	a -= b;
@@ -963,7 +1075,8 @@ static s32 s32_gte(s32 a, s32 b)
 /* Take a fully reduced polynomial form number and contract it into a
  * little-endian, 32-byte array.
  *
- * On entry: |input_limbs[i]| < 2^26 */
+ * On entry: |input_limbs[i]| < 2^26
+ */
 static void fcontract(u8 *output, limb *input_limbs)
 {
 	int i;
@@ -980,31 +1093,37 @@ static void fcontract(u8 *output, limb *input_limbs)
 		for (i = 0; i < 9; ++i) {
 			if ((i & 1) == 1) {
 				/* This calculation is a time-invariant way to make input[i]
-				 * non-negative by borrowing from the next-larger limb. */
+				 * non-negative by borrowing from the next-larger limb.
+				 */
 				const s32 mask = input[i] >> 31;
 				const s32 carry = -((input[i] & mask) >> 25);
+
 				input[i] = input[i] + (carry << 25);
 				input[i+1] = input[i+1] - carry;
 			} else {
 				const s32 mask = input[i] >> 31;
 				const s32 carry = -((input[i] & mask) >> 26);
+
 				input[i] = input[i] + (carry << 26);
 				input[i+1] = input[i+1] - carry;
 			}
 		}
 
 		/* There's no greater limb for input[9] to borrow from, but we can multiply
-		 * by 19 and borrow from input[0], which is valid mod 2^255-19. */
+		 * by 19 and borrow from input[0], which is valid mod 2^255-19.
+		 */
 		{
 			const s32 mask = input[9] >> 31;
 			const s32 carry = -((input[9] & mask) >> 25);
+
 			input[9] = input[9] + (carry << 25);
 			input[0] = input[0] - (carry * 19);
 		}
 
 		/* After the first iteration, input[1..9] are non-negative and fit within
 		 * 25 or 26 bits, depending on position. However, input[0] may be
-		 * negative. */
+		 * negative.
+		 */
 	}
 
 	/* The first borrow-propagation pass above ended with every limb
@@ -1020,20 +1139,24 @@ static void fcontract(u8 *output, limb *input_limbs)
 	{
 		const s32 mask = input[0] >> 31;
 		const s32 carry = -((input[0] & mask) >> 26);
+
 		input[0] = input[0] + (carry << 26);
 		input[1] = input[1] - carry;
 	}
 
 	/* All input[i] are now non-negative. However, there might be values between
-	 * 2^25 and 2^26 in a limb which is, nominally, 25 bits wide. */
+	 * 2^25 and 2^26 in a limb which is, nominally, 25 bits wide.
+	 */
 	for (j = 0; j < 2; j++) {
 		for (i = 0; i < 9; i++) {
 			if ((i & 1) == 1) {
 				const s32 carry = input[i] >> 25;
+
 				input[i] &= 0x1ffffff;
 				input[i+1] += carry;
 			} else {
 				const s32 carry = input[i] >> 26;
+
 				input[i] &= 0x3ffffff;
 				input[i+1] += carry;
 			}
@@ -1041,6 +1164,7 @@ static void fcontract(u8 *output, limb *input_limbs)
 
 		{
 			const s32 carry = input[9] >> 25;
+
 			input[9] &= 0x1ffffff;
 			input[0] += 19*carry;
 		}
@@ -1051,11 +1175,13 @@ static void fcontract(u8 *output, limb *input_limbs)
 	 * < 2^26 + 2*19, because the carry was, at most, two.
 	 *
 	 * If the second pass carried from input[9] again then input[0] is < 2*19 and
-	 * the input[9] -> input[0] carry didn't push input[0] out of bounds. */
+	 * the input[9] -> input[0] carry didn't push input[0] out of bounds.
+	 */
 
 	/* It still remains the case that input might be between 2^255-19 and 2^255.
 	 * In this case, input[1..9] must take their maximum value and input[0] must
-	 * be >= (2^255-19) & 0x3ffffff, which is 0x3ffffed. */
+	 * be >= (2^255-19) & 0x3ffffff, which is 0x3ffffed.
+	 */
 	mask = s32_gte(input[0], 0x3ffffed);
 	for (i = 1; i < 10; i++) {
 		if ((i & 1) == 1) {
@@ -1066,7 +1192,8 @@ static void fcontract(u8 *output, limb *input_limbs)
 	}
 
 	/* mask is either 0xffffffff (if input >= 2^255-19) and zero otherwise. Thus
-	 * this conditionally subtracts 2^255-19. */
+	 * this conditionally subtracts 2^255-19.
+	 */
 	input[0] -= mask & 0x3ffffed;
 
 	for (i = 1; i < 10; i++) {
@@ -1092,16 +1219,16 @@ static void fcontract(u8 *output, limb *input_limbs)
 	output[s+3]  = (input[i] >> 24) & 0xff;
 	output[0] = 0;
 	output[16] = 0;
-	F(0,0);
-	F(1,3);
-	F(2,6);
-	F(3,9);
-	F(4,12);
-	F(5,16);
-	F(6,19);
-	F(7,22);
-	F(8,25);
-	F(9,28);
+	F(0, 0);
+	F(1, 3);
+	F(2, 6);
+	F(3, 9);
+	F(4, 12);
+	F(5, 16);
+	F(6, 19);
+	F(7, 22);
+	F(8, 25);
+	F(9, 28);
 #undef F
 }
 
@@ -1113,14 +1240,16 @@ static void fcontract(u8 *output, limb *input_limbs)
  * wrong results.  Also, the two limb arrays must be in reduced-coefficient,
  * reduced-degree form: the values in a[10..19] or b[10..19] aren't swapped,
  * and all all values in a[0..9],b[0..9] must have magnitude less than
- * INT32_MAX. */
+ * INT32_MAX.
+ */
 static void swap_conditional(limb a[19], limb b[19], limb iswap)
 {
-	unsigned i;
+	unsigned int i;
 	const s32 swap = (s32) -iswap;
 
 	for (i = 0; i < 10; ++i) {
-		const s32 x = swap & ( ((s32)a[i]) ^ ((s32)b[i]) );
+		const s32 x = swap & (((s32)a[i]) ^ ((s32)b[i]));
+
 		a[i] = ((s32)a[i]) ^ x;
 		b[i] = ((s32)b[i]) ^ x;
 	}
@@ -1140,57 +1269,57 @@ static void crecip(limb *out, const limb *z)
 	limb t1[10];
 	int i;
 
-	/* 2 */ fsquare(z2,z);
-	/* 4 */ fsquare(t1,z2);
-	/* 8 */ fsquare(t0,t1);
-	/* 9 */ fmul(z9,t0,z);
-	/* 11 */ fmul(z11,z9,z2);
-	/* 22 */ fsquare(t0,z11);
-	/* 2^5 - 2^0 = 31 */ fmul(z2_5_0,t0,z9);
+	/* 2 */ fsquare(z2, z);
+	/* 4 */ fsquare(t1, z2);
+	/* 8 */ fsquare(t0, t1);
+	/* 9 */ fmul(z9, t0, z);
+	/* 11 */ fmul(z11, z9, z2);
+	/* 22 */ fsquare(t0, z11);
+	/* 2^5 - 2^0 = 31 */ fmul(z2_5_0, t0, z9);
 
-	/* 2^6 - 2^1 */ fsquare(t0,z2_5_0);
-	/* 2^7 - 2^2 */ fsquare(t1,t0);
-	/* 2^8 - 2^3 */ fsquare(t0,t1);
-	/* 2^9 - 2^4 */ fsquare(t1,t0);
-	/* 2^10 - 2^5 */ fsquare(t0,t1);
-	/* 2^10 - 2^0 */ fmul(z2_10_0,t0,z2_5_0);
+	/* 2^6 - 2^1 */ fsquare(t0, z2_5_0);
+	/* 2^7 - 2^2 */ fsquare(t1, t0);
+	/* 2^8 - 2^3 */ fsquare(t0, t1);
+	/* 2^9 - 2^4 */ fsquare(t1, t0);
+	/* 2^10 - 2^5 */ fsquare(t0, t1);
+	/* 2^10 - 2^0 */ fmul(z2_10_0, t0, z2_5_0);
 
-	/* 2^11 - 2^1 */ fsquare(t0,z2_10_0);
-	/* 2^12 - 2^2 */ fsquare(t1,t0);
-	/* 2^20 - 2^10 */ for (i = 2; i < 10; i += 2) { fsquare(t0,t1); fsquare(t1,t0); }
-	/* 2^20 - 2^0 */ fmul(z2_20_0,t1,z2_10_0);
+	/* 2^11 - 2^1 */ fsquare(t0, z2_10_0);
+	/* 2^12 - 2^2 */ fsquare(t1, t0);
+	/* 2^20 - 2^10 */ for (i = 2; i < 10; i += 2) { fsquare(t0, t1); fsquare(t1, t0); }
+	/* 2^20 - 2^0 */ fmul(z2_20_0, t1, z2_10_0);
 
-	/* 2^21 - 2^1 */ fsquare(t0,z2_20_0);
-	/* 2^22 - 2^2 */ fsquare(t1,t0);
-	/* 2^40 - 2^20 */ for (i = 2; i < 20; i += 2) { fsquare(t0,t1); fsquare(t1,t0); }
-	/* 2^40 - 2^0 */ fmul(t0,t1,z2_20_0);
+	/* 2^21 - 2^1 */ fsquare(t0, z2_20_0);
+	/* 2^22 - 2^2 */ fsquare(t1, t0);
+	/* 2^40 - 2^20 */ for (i = 2; i < 20; i += 2) { fsquare(t0, t1); fsquare(t1, t0); }
+	/* 2^40 - 2^0 */ fmul(t0, t1, z2_20_0);
 
-	/* 2^41 - 2^1 */ fsquare(t1,t0);
-	/* 2^42 - 2^2 */ fsquare(t0,t1);
-	/* 2^50 - 2^10 */ for (i = 2; i < 10; i += 2) { fsquare(t1,t0); fsquare(t0,t1); }
-	/* 2^50 - 2^0 */ fmul(z2_50_0,t0,z2_10_0);
+	/* 2^41 - 2^1 */ fsquare(t1, t0);
+	/* 2^42 - 2^2 */ fsquare(t0, t1);
+	/* 2^50 - 2^10 */ for (i = 2; i < 10; i += 2) { fsquare(t1, t0); fsquare(t0, t1); }
+	/* 2^50 - 2^0 */ fmul(z2_50_0, t0, z2_10_0);
 
-	/* 2^51 - 2^1 */ fsquare(t0,z2_50_0);
-	/* 2^52 - 2^2 */ fsquare(t1,t0);
-	/* 2^100 - 2^50 */ for (i = 2; i < 50; i += 2) { fsquare(t0,t1); fsquare(t1,t0); }
-	/* 2^100 - 2^0 */ fmul(z2_100_0,t1,z2_50_0);
+	/* 2^51 - 2^1 */ fsquare(t0, z2_50_0);
+	/* 2^52 - 2^2 */ fsquare(t1, t0);
+	/* 2^100 - 2^50 */ for (i = 2; i < 50; i += 2) { fsquare(t0, t1); fsquare(t1, t0); }
+	/* 2^100 - 2^0 */ fmul(z2_100_0, t1, z2_50_0);
 
-	/* 2^101 - 2^1 */ fsquare(t1,z2_100_0);
-	/* 2^102 - 2^2 */ fsquare(t0,t1);
-	/* 2^200 - 2^100 */ for (i = 2; i < 100; i += 2) { fsquare(t1,t0); fsquare(t0,t1); }
-	/* 2^200 - 2^0 */ fmul(t1,t0,z2_100_0);
+	/* 2^101 - 2^1 */ fsquare(t1, z2_100_0);
+	/* 2^102 - 2^2 */ fsquare(t0, t1);
+	/* 2^200 - 2^100 */ for (i = 2; i < 100; i += 2) { fsquare(t1, t0); fsquare(t0, t1); }
+	/* 2^200 - 2^0 */ fmul(t1, t0, z2_100_0);
 
-	/* 2^201 - 2^1 */ fsquare(t0,t1);
-	/* 2^202 - 2^2 */ fsquare(t1,t0);
-	/* 2^250 - 2^50 */ for (i = 2; i < 50; i += 2) { fsquare(t0,t1); fsquare(t1,t0); }
-	/* 2^250 - 2^0 */ fmul(t0,t1,z2_50_0);
+	/* 2^201 - 2^1 */ fsquare(t0, t1);
+	/* 2^202 - 2^2 */ fsquare(t1, t0);
+	/* 2^250 - 2^50 */ for (i = 2; i < 50; i += 2) { fsquare(t0, t1); fsquare(t1, t0); }
+	/* 2^250 - 2^0 */ fmul(t0, t1, z2_50_0);
 
-	/* 2^251 - 2^1 */ fsquare(t1,t0);
-	/* 2^252 - 2^2 */ fsquare(t0,t1);
-	/* 2^253 - 2^3 */ fsquare(t1,t0);
-	/* 2^254 - 2^4 */ fsquare(t0,t1);
-	/* 2^255 - 2^5 */ fsquare(t1,t0);
-	/* 2^255 - 21 */ fmul(out,t1,z11);
+	/* 2^251 - 2^1 */ fsquare(t1, t0);
+	/* 2^252 - 2^2 */ fsquare(t0, t1);
+	/* 2^253 - 2^3 */ fsquare(t1, t0);
+	/* 2^254 - 2^4 */ fsquare(t0, t1);
+	/* 2^255 - 2^5 */ fsquare(t1, t0);
+	/* 2^255 - 21 */ fmul(out, t1, z11);
 }
 
 
@@ -1205,11 +1334,13 @@ static void crecip(limb *out, const limb *z)
  *   qmqp: short form, preserved
  *
  * On entry and exit, the absolute value of the limbs of all inputs and outputs
- * are < 2^26. */
+ * are < 2^26.
+ */
 static void fmonty(limb *x2, limb *z2,  /* output 2Q */
 		   limb *x3, limb *z3,  /* output Q + Q' */
 		   limb *x, limb *z,    /* input Q */
 		   limb *xprime, limb *zprime,  /* input Q' */
+
 		   const limb *qmqp /* input Q - Q' */)
 {
 	limb origx[10], origxprime[10], zzz[19], xx[19], zz[19], xxprime[19],
@@ -1229,7 +1360,8 @@ static void fmonty(limb *x2, limb *z2,  /* output 2Q */
 	fproduct(xxprime, xprime, z);
 	/* |xxprime[i]| < 14*2^54: the largest product of two limbs will be <
 	 * 2^(27+27) and fproduct adds together, at most, 14 of those products.
-	 * (Approximating that to 2^58 doesn't work out.) */
+	 * (Approximating that to 2^58 doesn't work out.)
+	 */
 	fproduct(zzprime, x, zprime);
 	/* |zzprime[i]| < 14*2^54 */
 	freduce_degree(xxprime);
@@ -1286,7 +1418,8 @@ static void fmonty(limb *x2, limb *z2,  /* output 2Q */
  *
  *   resultx/resultz: the x coordinate of the resulting curve point (short form)
  *   n: a little endian, 32-byte number
- *   q: a point of the curve (short form) */
+ *   q: a point of the curve (short form)
+ */
 static void cmult(limb *resultx, limb *resultz, const u8 *n, const limb *q)
 {
 	limb a[19] = {0}, b[19] = {1}, c[19] = {1}, d[19] = {0};
@@ -1294,12 +1427,13 @@ static void cmult(limb *resultx, limb *resultz, const u8 *n, const limb *q)
 	limb e[19] = {0}, f[19] = {1}, g[19] = {0}, h[19] = {1};
 	limb *nqpqx2 = e, *nqpqz2 = f, *nqx2 = g, *nqz2 = h;
 
-	unsigned i, j;
+	unsigned int i, j;
 
 	memcpy(nqpqx, q, sizeof(limb) * 10);
 
 	for (i = 0; i < 32; ++i) {
 		u8 byte = n[31 - i];
+
 		for (j = 0; j < 8; ++j) {
 			const limb bit = byte >> 7;
 
@@ -1336,7 +1470,7 @@ static void cmult(limb *resultx, limb *resultz, const u8 *n, const limb *q)
 
 bool curve25519(u8 mypublic[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_POINT_SIZE], const u8 basepoint[CURVE25519_POINT_SIZE])
 {
-#if IS_ENABLED(CONFIG_KERNEL_MODE_NEON) && defined(CONFIG_ARM)
+#if IS_ENABLED(CONFIG_KERNEL_MODE_NEON) && defined(CONFIG_ARM) && !defined(CONFIG_CPU_THUMBONLY)
 	if (curve25519_use_neon && may_use_simd()) {
 		kernel_neon_begin();
 		curve25519_asm_neon(mypublic, secret, basepoint);
@@ -1382,12 +1516,14 @@ struct other_stack {
  *   qmqp: short form, preserved
  *
  * On entry and exit, the absolute value of the limbs of all inputs and outputs
- * are < 2^26. */
+ * are < 2^26.
+ */
 static void fmonty(struct other_stack *s,
 		   limb *x2, limb *z2,  /* output 2Q */
 		   limb *x3, limb *z3,  /* output Q + Q' */
 		   limb *x, limb *z,    /* input Q */
 		   limb *xprime, limb *zprime,  /* input Q' */
+
 		   const limb *qmqp /* input Q - Q' */)
 {
 	memcpy(s->origx, x, 10 * sizeof(limb));
@@ -1404,7 +1540,8 @@ static void fmonty(struct other_stack *s,
 	fproduct(s->xxprime, xprime, z);
 	/* |s->xxprime[i]| < 14*2^54: the largest product of two limbs will be <
 	 * 2^(27+27) and fproduct adds together, at most, 14 of those products.
-	 * (Approximating that to 2^58 doesn't work out.) */
+	 * (Approximating that to 2^58 doesn't work out.)
+	 */
 	fproduct(s->zzprime, x, zprime);
 	/* |s->zzprime[i]| < 14*2^54 */
 	freduce_degree(s->xxprime);
@@ -1461,10 +1598,11 @@ static void fmonty(struct other_stack *s,
  *
  *   resultx/resultz: the x coordinate of the resulting curve point (short form)
  *   n: a little endian, 32-byte number
- *   q: a point of the curve (short form) */
+ *   q: a point of the curve (short form)
+ */
 static void cmult(struct other_stack *s, limb *resultx, limb *resultz, const u8 *n, const limb *q)
 {
-	unsigned i, j;
+	unsigned int i, j;
 	limb *nqpqx = s->a, *nqpqz = s->b, *nqx = s->c, *nqz = s->d, *t;
 	limb *nqpqx2 = s->e, *nqpqz2 = s->f, *nqx2 = s->g, *nqz2 = s->h;
 
@@ -1473,6 +1611,7 @@ static void cmult(struct other_stack *s, limb *resultx, limb *resultz, const u8
 
 	for (i = 0; i < 32; ++i) {
 		u8 byte = n[31 - i];
+
 		for (j = 0; j < 8; ++j) {
 			const limb bit = byte >> 7;
 
@@ -1510,7 +1649,7 @@ static void cmult(struct other_stack *s, limb *resultx, limb *resultz, const u8
 
 bool curve25519(u8 mypublic[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_POINT_SIZE], const u8 basepoint[CURVE25519_POINT_SIZE])
 {
-#if IS_ENABLED(CONFIG_KERNEL_MODE_NEON) && defined(CONFIG_ARM)
+#if IS_ENABLED(CONFIG_KERNEL_MODE_NEON) && defined(CONFIG_ARM) && !defined(CONFIG_CPU_THUMBONLY)
 	if (curve25519_use_neon && may_use_simd()) {
 		kernel_neon_begin();
 		curve25519_asm_neon(mypublic, secret, basepoint);
@@ -1519,6 +1658,7 @@ bool curve25519(u8 mypublic[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_P
 #endif
 	{
 		struct other_stack *s = kzalloc(sizeof(struct other_stack), GFP_KERNEL);
+
 		if (unlikely(!s))
 			return false;
 
@@ -1539,13 +1679,17 @@ bool curve25519(u8 mypublic[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_P
 bool curve25519_generate_public(u8 pub[CURVE25519_POINT_SIZE], const u8 secret[CURVE25519_POINT_SIZE])
 {
 	static const u8 basepoint[CURVE25519_POINT_SIZE] __aligned(32) = { 9 };
+
+	if (unlikely(!crypto_memneq(secret, null_point, CURVE25519_POINT_SIZE)))
+		return false;
+
 	return curve25519(pub, secret, basepoint);
 }
 #endif
 
 void curve25519_generate_secret(u8 secret[CURVE25519_POINT_SIZE])
 {
-	get_random_bytes(secret, CURVE25519_POINT_SIZE);
+	get_random_bytes_wait(secret, CURVE25519_POINT_SIZE);
 	normalize_secret(secret);
 }
 
